/* -*-Mode: C++;-*-
 * PRCS - The Project Revision Control System
 * Copyright (C) 1997  Josh MacDonald
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: vc.h 1.6.1.5.1.11.1.1.1.2 Wed, 22 Jan 1997 12:02:00 -0800 jmacd $
 */


#ifndef _VC_H_
#define _VC_H_

#include "utils.h"
#include "config.h"

EXTERN int VC_check_token_match(const char* token, const char* type);

#ifdef __cplusplus    /* the C flex parser includes this and doesn't like
                         some of the C++ below (especially the include) */

PrVoidError VC_register(const char* newfile);

PrVoidError VC_checkin(const char* workingfile,
		       const char* parentversion,
		       const char* revisionfile,
		       const char* log,
		       void* data,
		       void (* notify_func)(void* data, const char* new_version),
		       bool keep);

PrCFilePtrError VC_checkout_stream(const char* version,
				   const char* revisionfile);

PrVoidError VC_close_checkout_stream(FILE* stream, const char* version,
				     const char* revisionfile);

PrVoidError VC_checkout_file(const char* newfile,
			     const char* version,
			     const char* revisionfile);

PrBoolError VC_delete_version(const char* version,
			      const char* revisionfile);

PrVoidError VC_set_log(const char* new_log_text,
		       const char* version,
		       const char* revisionfile);

/*
 * RcsVersionData --
 *
 *     This data structure stores all the important information about
 *     an RCS delta as would be extracted from its log.  It contains
 *     the time it was checked in(seconds since the epoch), the
 *     author("jmacd"), the MD5 digest, and its version string("1.2") */
class RcsVersionData {
public:
    enum GCMark { Referenced, NotReferenced, ReferencedTraversed };

    RcsVersionData();

    time_t date() const;
    int length() const;
    const char* author() const;
    const char* rcs_version() const;
    const char* unkeyed_checksum() const;
    GCMark referenced() const;

    void date(time_t t0);
    void length(int l0);
    void author(const char* a0);
    void rcs_version(const char* rv0);
    void unkeyed_checksum(const char* ck0);
    void reference(GCMark gc0);

    virtual bool OK() const;

protected:
    time_t _date;
    const char* _author;
    const char* _rcs_version;
    char _unkeyed_checksum[16];
    int _length;
    GCMark _gc_mark;
};

/*
 * RcsVersionData --
 *
 *     This data structure stores all the important information about
 *     an project version file's RCS delta.  It will always have the
 *     following data to be considered OK().
 *
 *     deleted -- whether this version has been deleted.  The delta remains
 *     for the purposes of tracking its history.
 *
 *     PrcsMajor, PrcsMinor, ParentMajor, ParentMinor -- strings representing
 *     the various version numbers.
 *
 *     The following members may or may not be present:
 *
 *     DescendsFromMajor, DescendsFromMinor -- If a branch was collapsed
 *     at checkin into a new major branch (when its root in the new branch
 *     is the same as the new parent), this is the "correct" parent.
 *
 *     MergeParentMajors, MergeParentMinors -- If this version was merged
 *     with another, these represent the history of "correct" parents.  If
 *     version A.3 with parent A.2 was merged with B.4 with parent B.3
 *     and checked in, the version's parent will be B.4, and the first
 *     MergeParent will be A.3.
 */
class ProjectVersionData : public RcsVersionData {
public:
    ProjectVersionData();

    bool deleted() const;
    const char* prcs_major() const;
    const char* prcs_minor() const;
    int prcs_minor_int() const;
    const char* parent_major() const;
    const char* parent_minor() const;
    int parent_minor_int() const;
    const char* descends_from_major() const;
    const char* descends_from_minor() const;
    bool has_descends_from() const;

    void deleted(bool del0);
    void prcs_major(const char* pm0);
    void prcs_minor(const char* pm0);
    void parent_major(const char* pm0);
    void parent_minor(const char* pm0);
    void descends_from_major(const char* df0);
    void descends_from_minor(const char* df0);

    virtual bool OK() const;
protected:
    bool _deleted;
    const char* _prcs_major;
    const char* _prcs_minor;
    int _prcs_minor_int;
    const char* _parent_major;
    const char* _parent_minor;
    int _parent_minor_int;
    const char* _descends_from_major;
    const char* _descends_from_minor;
};

/*
 * RcsDelta --
 *
 *     RCS version numbers are organized as described in rcsfile(5).  These
 *     two classes organize a set of deltas in a particular file so that
 *     lookup of a version string "1.2.3.4" may be done with 4 lookups.
 *     A RcsBranch has a branch number such as 1 in the above example and
 *     an array of RcsDeltas.  A RcsDelta has a delta number such as 2 in
 *     the above example and an array of branches.  They are constructed
 *     for an entire RCS file with a call to VC_construct_delta_array.
 */
class RcsDelta {
public:
    RcsDelta(RcsDelta* parent0);

    ~RcsDelta();

    RcsVersionData* lookup(const char* index);
    void insert(const char* index, RcsVersionData* data);

    class DeltaIterator {
    public:
	DeltaIterator(RcsDelta* d0);
	RcsVersionData* operator*() const;
	void next();
	bool finished() const;
    private:
	RcsDelta *_current;
	RcsVersionData* _current_data;
	bool _finished;
	int _index;
    };

    void remove(const char* index);
    int count() const;

    friend class DeltaIterator;

private:

    RcsVersionData* rcs_data;
    RcsDeltaPtrArray* delta_array;
    int insertion_count;

    /* for use by DeltaIterator */
    RcsDelta *parent;
    int my_iter_index;
};

PrRcsDeltaPtrError
VC_get_version_data(const char* versionfile);

PrProjectVersionDataPtrArrayPtrError
VC_get_project_version_data(const char* versionfile);

#endif

enum RcsToken {
    NoTokenMatch = 0,            /* This is actually what yylex() returns */
    RcsAbort = 1,                /* This matches the aborted message */

    InitialRevision = 2,         /* Possible checkin outputs. */
    NewRevision = 3,
    PrevRevision = 4,

    RlogVersion = 5,             /* Possible rlog outputs. */
    RlogDate = 6,                /* RlogVersion also matches the co output */
    RlogAuthor = 7,

    PrcsMajorVersion = 8,
    PrcsMinorVersion = 9,
    PrcsParentMajorVersion = 10,
    PrcsParentMinorVersion = 11,

    PrcsVersionDeleted = 12,

    RlogTotalRevisions = 13,

    PrcsMergedParentMajor = 15,
    PrcsMergedParentMinor = 16,
    PrcsDescendsFromMajor = 17,
    PrcsDescendsFromMinor = 18
};

#endif
