/* -*-Mode: C++;-*-
 * PRCS - The Project Revision Control System
 * Copyright (C) 1997  Josh MacDonald
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: merge.cc 1.18.1.1.1.26.1.1.1.4.1.9.1.11 Thu, 13 Feb 1997 00:48:05 -0800 jmacd $
 */

#include "prcs.h"
#include "hash.h"
#include "diff.h"
#include "populate.h"
#include "checkin.h"
#include "checkout.h"
#include "repository.h"
#include "projdesc.h"
#include "vc.h"
#include "syscmd.h"
#include "setkeys.h"
#include "fileent.h"
#include "misc.h"
#include "system.h"


static PrVoidError merge_common_files(ProjectDescriptor*,
				      ProjectDescriptor*,
				      ProjectDescriptor*,
				      FileTable*,
				      FileTable*,
				      FileTable*,
				      FileTable*,
				      FileTable*,
				      FileTable*);
static PrVoidError merge_selected_only_files(ProjectDescriptor*,
					     ProjectDescriptor*,
					     ProjectDescriptor*,
					     FileTable*,
					     FileTable*,
					     FileTable*,
					     FileTable*,
					     FileTable*,
					     FileTable*);
static PrVoidError merge_working_only_files(ProjectDescriptor*,
					    ProjectDescriptor*,
					    ProjectDescriptor*,
					    FileTable*,
					    FileTable*,
					    FileTable*,
					    FileTable*,
					    FileTable*,
					    FileTable*);
/*
 * 7 possible cases are handled:
 */
static PrVoidError merge_common_descendants(ProjectDescriptor*,
					    FileEntry*, FileEntry*, FileEntry*);
static PrVoidError merge_working_descendant(ProjectDescriptor*,
					    FileEntry*, FileEntry*, FileEntry*);
static PrVoidError merge_selected_descendant(ProjectDescriptor*,
					    FileEntry*, FileEntry*, FileEntry*);
static PrVoidError merge_selected_only_file(ProjectDescriptor*,
					    FileEntry*, FileEntry*, FileEntry*);
static PrVoidError merge_work_only_file(ProjectDescriptor*,
					FileEntry*, FileEntry*, FileEntry*);
static PrVoidError merge_no_descendant(ProjectDescriptor*,
				       FileEntry*, FileEntry*, FileEntry*);
static PrVoidError merge_no_ancestor(ProjectDescriptor*,
				     FileEntry*, FileEntry*, FileEntry*);

static PrVoidError cant_add(FileEntry* common_fe, FileEntry* selected_fe);
static PrVoidError merge_finish(MergeAction def, const char* valid, const char* detail,
				ProjectDescriptor*, FileEntry* selected,
				FileEntry* common, FileEntry* working);

static PrBoolError merge_files(ProjectDescriptor*, FileEntry*, FileEntry*, FileEntry*);
static PrVoidError replace_file(ProjectDescriptor*, FileEntry*, FileEntry*, FileEntry*);
static PrVoidError add_file(ProjectDescriptor*, FileEntry*, FileEntry*, FileEntry*);
static PrVoidError delete_file(ProjectDescriptor*, FileEntry*, FileEntry*, FileEntry*);

static PrBoolError cmp_filenames(const char*, const char*);
static PrVoidError checkout_unkey(const char* filename,
				  FileEntry*,
				  bool working,
				  RepEntry*,
				  ProjectDescriptor* );
static PrVoidError merge_cleanup_handler(void* data, bool);
static PrProjectDescriptorPtrError checkout_common_version(ProjectVersionData*,
							   ProjectVersionData*,
							   ProjectDescriptor*,
							   RepEntry* rep_entry,
							   ProjectVersionData**);
static PrVoidError make_obsolete(const char* work_path);
static bool quick_compare(FileEntry* one,
			  FileEntry* two,
			  bool two_is_working);

static const char* selected_format;
static const char* common_format;
static const char* working_format;

#define temp_file_working temp_file_1
#define temp_file_common temp_file_2
#define temp_file_selected temp_file_3

static const char working_absent_sel_changed[] =
"The file has been deleted from your working version of the project, "
"and still exists in the selected version.  The selected version has "
"been modified, but it is still assumed that the file is obsolete";

static const char working_absent_sel_unchanged[] =
"The file has been deleted from your working version of the project, "
"and still exists in the selected version.  The selected version has "
"not been modified, so it is assumed that the file is obsolete";

static const char all_files_differ[] =
"There are no equivalent files among the three files in each version.  "
"It is assumed that you are interested in merging the selected "
"version's changes with your changes.  Therefore, the default action is "
"merge";

static const char just_selected_differs[] =
"Your version of the file is unchanged from the common version, yet "
"the selected version has been modified.  It is assumed then, that the "
"selected version of the file is more up to date, and the default "
"action is to replace your file with the file from the selected "
"version";

static const char selected_only_detail[] =
"The file exists only in the selected version.	Niether the common "
"version or your working version of the project contain the file.  The "
"default action is to add the file";

static const char no_ancestor_detail[] =
"The file exists in both the selected version and your version, but "
"not in the common version.  The files differ.  The default action is "
"to merge, with an empty file used as the common version.  This will "
"produce serious conflicts";

static const char selected_file_absent[] =
"The file has been deleted from the selected of the project, and still "
"exists in your version.  You have not modified the file, so your version "
"is assumed to be obsolete.  Therefore, the default action is delete";

static const char ambig_help_string[] =
"You have merged the working project against at least one version prior "
"to this merge.  PRCS must compute the common version between the "
"selected version and the effective working version, but the effective "
"working version is ambiguous because the working project has already "
"been merged once.  Picking a version here is complicated, so instead "
"of making the choice automatically and having someone complain that it "
"made the wrong choice you have been given complete control";

static const char cant_add_help_string[] =
"The file does not exist in your working project and may or may "
"not exist in the common project version.  Either way, the file "
"you would be adding will conflict with another file already in the "
"project.  Normally this does not happen, but when you've done a previous "
"merge with a different setting of the -F (file family) option, you "
"can cause this situation by renaming a file in a previous version, "
"merging with -F, and now trying to add the old name back again by "
"merging without -F.  The opposite case can occur too.  There is nothing "
"you can do with this file, you must manually correct this conflict.  "
"It may be the case that you have just deleted or are about to delete "
"the conflicting file entry in the working project, but since it would "
"make merges dependent on the order files are treated it is figured "
"that if you got yourself into this situation you probably know how best "
"to get yourself out.  For this reason, merge does not handle possible "
"adds last, after all possible deletions have been done.  You can "
"correct this problem by re-merging with the -s (--skilled-merge) option";

PrPrcsExitStatusError merge_command()
{
    FileTable common_files(pathname_hash, pathname_equal),
	      work_files(pathname_hash, pathname_equal),
	      selected_files(pathname_hash, pathname_equal);
    FileTable work_files_alt(pathname_hash, pathname_equal);
    ProjectDescriptor *Working, *Selected, *Common;
    ProjectVersionData *selected_version, *working_version, *eff_working_version;
    RepEntry* rep_entry;
    Dstring selfmt, comfmt, wrkfmt;

    Return_if_fail(Working << read_project_file(cmd_root_project_full_name,
						cmd_root_project_file_path,
						(ProjectReadData)(KeepMergeParents | KeepKeywords)));

    Return_if_fail(Working->init_repository_entry(cmd_root_project_name, false, false));

    rep_entry = Working->repository_entry();

    working_version = rep_entry->lookup_version(*Working->project_version_major(),
						*Working->project_version_minor());

    if(!working_version)
	throw prcserror << "Invalid version in working project file, cannot merge" << dotendl;

    eliminate_unnamed_files(Working);

    Return_if_fail(eliminate_working_files(Working, SetNotPresentWarnUser));

    Working->read_quick_elim();
    Working->quick_elim_unmodified();

    Return_if_fail(selected_version << resolve_version(cmd_version_specifier_major,
						       cmd_version_specifier_minor,
						       cmd_root_project_full_name,
						       cmd_root_project_file_path,
						       Working,
						       rep_entry));

    if(selected_version->prcs_minor_int() == 0) {
 	Return_if_fail(Selected << checkout_empty_prj_file(cmd_root_project_full_name,
							   selected_version->prcs_major(),
							   (ProjectReadData)(KeepMergeParents | KeepKeywords)));
    } else {
	Return_if_fail(Selected << rep_entry->checkout_prj_file(cmd_root_project_full_name,
								selected_version->rcs_version(),
								(ProjectReadData)(KeepMergeParents | KeepKeywords)));
    }

    selfmt.sprintf("%s.%s", selected_version->prcs_major(),
		   selected_version->prcs_minor());
    selected_format = selfmt;

    prcsinfo << "Selected version: " << selected_format << prcsendl;

    Return_if_fail(Common << checkout_common_version(working_version, selected_version,
						     Working, rep_entry, &eff_working_version));

    if (!Common) { /* merge is complete. */
	prcsinfo << "Selected version has already been completely merged against"
		 << dotendl;
	return ExitSuccess;
    }

    eliminate_unnamed_files(Selected);
    eliminate_unnamed_files(Common);

    Common->repository_entry(Working->repository_entry());
    Selected->repository_entry(Working->repository_entry());

    Return_if_fail(warn_unused_files(true));

    comfmt.sprintf("%s.%s", Common->project_version_major()->cast(),
		   Common->project_version_minor()->cast());
    common_format = comfmt;

    wrkfmt.sprintf("%s.%s", eff_working_version->prcs_major(),
		   eff_working_version->prcs_minor());
    working_format = wrkfmt;

    prcsinfo << "Common version:   " << common_format << prcsendl;
    prcsinfo << "Working files:    " << working_format << prcsendl;

    if (Working->merge_continuing (selected_version, eff_working_version))
        prcsinfo << "Continuing in progress merge" << dotendl;

    populate_tables(Common, &common_files, NULL);
    populate_tables(Working, &work_files, &work_files_alt);
    populate_tables(Selected, &selected_files, NULL);

    install_cleanup_handler(merge_cleanup_handler, Working);

    Return_if_fail(merge_common_files(Selected, Common, Working,
				      &selected_files, &common_files, &work_files,
				      NULL, NULL, &work_files_alt));

    Return_if_fail(merge_selected_only_files(Selected, Common, Working,
					     &selected_files, &common_files, &work_files,
					     NULL, NULL, &work_files_alt));

    Return_if_fail(merge_working_only_files(Selected, Common, Working,
					    &selected_files, &common_files, &work_files,
					    NULL, NULL, &work_files_alt));

    /* Note: merged project file is written out by merge_cleanup_handler */

    Working->merge_notify_complete();

    return ExitSuccess;
}

static MergeCandidate* common_merge_pred (MergeCandidate* mc,
					  RepEntry *rep_entry,
					  ProjectVersionData* selected,
					  MergeCandidate* closest,
					  bool* closest_ambig)
{
    mc->_common = rep_entry->common_version(selected, mc->_data, &mc->_common_distance);

    if (!mc->_common)
	return closest;

    if (closest && mc->_common_distance == closest->_common_distance &&
	(strcmp (mc->_data->prcs_major(), closest->_data->prcs_major()) != 0 ||
	 mc->_data->prcs_minor_int() != closest->_data->prcs_minor_int()))
	*closest_ambig = true;
    else if (!closest || mc->_common_distance < closest->_common_distance)
	return mc;

    return closest;
}

static bool merge_help_is_complete(MergeParentEntryPtrArray* entries, const char* selected_version)
{
    foreach (ent_ptr, entries, MergeParentEntryPtrArray::ArrayIterator) {
	MergeParentEntry *entry = *ent_ptr;

	if (entry->complete && strcmp (entry->selected_version, selected_version) == 0)
	    return true;
    }

    return false;
}

static PrVoidError merge_help_query_complete()
{
    prcsquery << "Selected version has already been completely merged against.  "
	      << force ("Aborting")
	      << report ("Abort")
	      << option ('y', "Continue anyway, you better know what you're doing")
	      << deffail ('n')
	      << query ("Continue");

    Return_if_fail (prcsquery.result());

    return NoError;
}

static PrVoidError merge_help_query_incomplete (const char* name)
{
    prcsquery << "Last merged against version, "
	      << name
	      << ", was not completely merged against.  "
	      << force("Aborting")
	      << report("Abort")
	      << option('y', "Continue with this merge")
	      << deffail('n')
	      << query("Continue");

    Return_if_fail (prcsquery.result());

    prcsinfo << "Assuming last version is completely merged against" << dotendl;

    return NoError;
}

static void merge_help_print_possible (MergeCandidate* cand)
{
    if (cand->_common_distance >= 0) {
	if (cand->_is_working) {
	    prcsinfo << "Working version ";
	} else {
	    prcsinfo << "Merge parent ";
	}

	prcsinfo << cand->_data->prcs_major() << '.' << cand->_data->prcs_minor()
		 << " with distance " << cand->_common_distance << prcsendl;
    }
}

/* Return whether files differ, possible false positive. */
static bool quick_compare(FileEntry* one,
			  FileEntry* two,
			  bool two_is_working)
{
    if (two_is_working && !two->unmodified() ||
	two_is_working && !two->descriptor_name())
	return true;

    return strcmp (one->descriptor_name(), two->descriptor_name()) != 0 ||
	   strcmp (one->descriptor_version_number(),
		   two->descriptor_version_number()) != 0;
}

static PrProjectDescriptorPtrError checkout_common_version(ProjectVersionData* working,
							   ProjectVersionData* selected,
							   ProjectDescriptor* project,
							   RepEntry* rep_entry,
							   ProjectVersionData** eff_working_version)
{
    ProjectVersionData *common = NULL;
    MergeParentEntryPtrArray* entries = project->new_merge_parents();
    int versions = entries->length();
    Dstring selected_version;

    selected_version.assign (selected->prcs_major());
    selected_version.append (".");
    selected_version.append (selected->prcs_minor());

    *eff_working_version = NULL;

    if (versions == 0) {
	*eff_working_version = working;
    } else {

	if (!entries->last_index()->complete &&
	    strcmp (entries->last_index()->selected_version, selected_version) != 0) {
	    /* Last merge is incomplete, ask to abort this one and finish last one. */

	    Return_if_fail (merge_help_query_incomplete(entries->last_index()->selected_version));
	}

	if (merge_help_is_complete(entries, selected_version)) {
	    /* Merge is complete already, if skilled ask to continue, otherwise
	     * just exit. */

	    if (option_skilled_merge) {
		Return_if_fail (merge_help_query_complete());
	    } else {
		return (ProjectDescriptor*) 0;
	    }
	}

	if (!option_skilled_merge && !entries->last_index()->complete) {
	    /* Continuing an in progress merge, select same effective working
	     * version as before. */

	    *eff_working_version = rep_entry->lookup_version
		(major_version_of (entries->last_index()->eff_working_version),
		 minor_version_of (entries->last_index()->eff_working_version));

	    if (!*eff_working_version)
		throw prcserror << "New-Merge-Parents entry version "
				<< entries->last_index()->eff_working_version
				<< " is not found in repository" << dotendl;
	} else /* versions > 0 */ {

	    /* Handle "The Marco Problem." */
	    /* In versions 0.x this defaulted to the last merge parent. */
	    /* In versions 1.0.x this defaulted to the working version. */
	    /* In versions 1.1 and later, if not skilled_merge, from the
	     * set containing all merge parents and the working project
	     * version, pick the version which has the closest distance
	     * to the common version which will be computed. */

	    MergeCandidatePtrArray *merge_candidates = new MergeCandidatePtrArray;

	    foreach (ent_ptr, entries, MergeParentEntryPtrArray::ArrayIterator) {
		MergeParentEntry *entry = *ent_ptr;
		ProjectVersionData *data;

		data = rep_entry->lookup_version (major_version_of (entry->selected_version),
						  minor_version_of (entry->selected_version));

		if (!data)
		    throw prcserror << "Named New-Merge-Parent version "
				    << entry->selected_version
				    << " not found in repository, run "
				    << squote ("prcs admin rebuild")
				    << " to diagnose this problem" << dotendl;

		merge_candidates->append (new MergeCandidate (data, false));
	    }

	    merge_candidates->append (new MergeCandidate (working, true));

	    /* See if we can find a unique closest common version. */
	    MergeCandidate *closest = NULL;
	    bool            closest_ambig = false;

	    foreach (mc_ptr, merge_candidates, MergeCandidatePtrArray::ArrayIterator)
		closest = common_merge_pred (*mc_ptr, rep_entry, selected, closest, &closest_ambig);

	    if (closest) {
		if (!option_skilled_merge && !closest_ambig) {
		    *eff_working_version = closest->_data;
		} else /* option_skilled_merge || closest_ambig */ {

		    Dstring closest_version;
		    closest_version.assign (closest->_data->prcs_major());
		    closest_version.append (".");
		    closest_version.append (closest->_data->prcs_minor());

		    if (closest_ambig) {
			prcsquery << "Effective working version used to compute common "
			    "version is ambiguous.  "
				  << force ("Aborting")
				  << report ("Abort")
				  << option ('n', "Proceed to select a version manually")
				  << deffail ('y')
				  << help (ambig_help_string)
				  << query ("Abort");

			Return_if_fail (prcsquery.result());
		    }

		    while (!*eff_working_version) {
			prcsinfo << "Possible effective working versions:" << dotendl;

			foreach (mc_ptr, merge_candidates, MergeCandidatePtrArray::ArrayIterator)
			    merge_help_print_possible (*mc_ptr);

			if (!closest_ambig) {
			    prcsquery << "Effective working version computed is "
				      << closest_version
				      << " with distance " << closest->_common_distance << ".  "
				      << definput (closest_version)
				      << "Please select an effective working version";
			} else {
			    prcsquery << "Please select an effective working version"
				      << definput ("");
			}

			prcsquery << string_query("");

			const char* result;

			Return_if_fail (result << prcsquery.string_result());

			const char* result_major = major_version_of (result);
			const char* result_minor = minor_version_of (result);

			if (!result_major) {
			    prcsinfo << "Illegal version name" << dotendl;
			    continue;
			}

			foreach (mc_ptr, merge_candidates, MergeCandidatePtrArray::ArrayIterator) {
			    MergeCandidate* mc = *mc_ptr;

			    if (strcmp (result_major, mc->_data->prcs_major()) == 0 &&
				strcmp (result_minor, mc->_data->prcs_minor()) == 0) {
				*eff_working_version = mc->_data;
				break;
			    }
			}
		    }
		}
	    } /* else eff_working_version still null */
	}
    }

    if (*eff_working_version) {
	common = rep_entry->common_version(*eff_working_version, selected, NULL);

	while(common && common->deleted()) {
	    prcsinfo << "Common version " << common->parent_major()
		     << "." << common->parent_minor()
		     << " was deleted, using its parent" << dotendl;

	    common = rep_entry->lookup_version(common->parent_major(),
					       common->parent_minor());

	    if(common && common->parent_minor_int() == 0) {
		prcsinfo << "All common versions were deleted, common version is empty" << dotendl;
		common = NULL;
	    }
	}
    }

    if(common == NULL) {
	prcsquery << "Common version is empty, you may use the empty version "
		  << working->prcs_major() << ".0" << ", but you better know what "
	    "you are doing"
		  << force ("Aborting")
		  << report ("Abort")
		  << option ('n', "Continue using empty version")
		  << deffail ('y')
		  << query ("Abort");

	Return_if_fail (prcsquery.result());

	return checkout_empty_prj_file(project->project_full_name(), working->prcs_major(),
				       (ProjectReadData)(KeepMergeParents | KeepKeywords));
    } else {
	return rep_entry->checkout_prj_file(project->project_full_name(), common->rcs_version(),
					    (ProjectReadData)(KeepMergeParents | KeepKeywords));
    }
}

static PrVoidError cant_add(FileEntry* common_fe, FileEntry* selected_fe)
{
    static BangFlag bang;

    prcsquery << "The file " << merge_tuple(NULL, common_fe, selected_fe)
	      << " exists in only the selected "
	      << (common_fe ? "and common project versions" : "project version")
	      << ".  The merge action would be to add or do nothing, but the "
	      << (option_file_family ? "file name" : "file family")
	      << " already exists in the working project.  "
	      << force ("No action taken")
	      << report ("No action will be taken")
	      << allow_bang (bang)
	      << help (cant_add_help_string)
	      << defopt ('c', "Continue")
	      << query ("No action will be taken, press a key to continue");

    Return_if_fail (prcsquery.result());

    return NoError;
}

static PrBoolError remerge_files (FileEntry *work,
				  FileEntry *common,
				  FileEntry *selected)
{
    static BangFlag remerge_bang;

    if (option_skilled_merge) {
	prcsquery << "The file " << merge_tuple(work, common, selected)
		  << " has already been merged against the selected version.  "
		  << force ("Remerging")
		  << report ("Remerge")
		  << allow_bang (remerge_bang)
		  << option ('n', "Ignore this file")
		  << defopt ('y', "Remerge this file")
		  << query ("Remerge");

	char c;

	Return_if_fail (c << prcsquery.result());

	if (c == 'y')
	    return true;
    }

    return false;
}

static PrVoidError merge_common_files(ProjectDescriptor* /* Selected */,
				      ProjectDescriptor* Common,
				      ProjectDescriptor* Working,
				      FileTable* selected_files,
				      FileTable* /* common_files */,
				      FileTable* work_files,
				      FileTable* /* selected_files_alt */,
				      FileTable* /* common_files_alt */,
				      FileTable* work_files_alt)
{
    FileEntry *common_fe, *selected_fe, *working_fe;

    foreach_fileent(fe_ptr, Common) {
	common_fe = *fe_ptr;

	working_fe = lookup_file(common_fe, work_files);
	selected_fe = lookup_file(common_fe, selected_files);

	bool was_merged = Working->has_been_merged(working_fe, common_fe, selected_fe);

	if(was_merged) {
	    bool remerge;
	    Return_if_fail (remerge << remerge_files (working_fe, common_fe, selected_fe));

	    if (!remerge)
		continue;
	}

	if(!was_merged &&
	   !common_fe->on_command_line() &&
	   (working_fe && !working_fe->on_command_line()) &&
	   (selected_fe && !selected_fe->on_command_line())) {

	    Working->merge_notify_incomplete();
	    continue;
	}

	if(working_fe) {

	    if(selected_fe) {
		Return_if_fail(merge_common_descendants(Working, selected_fe,
							common_fe, working_fe));
	    } else {
		Return_if_fail(merge_working_descendant(Working, NULL,
							common_fe, working_fe));
	    }

	} else if(selected_fe) {

	    if (alt_lookup_file (selected_fe, work_files_alt)) {
		Return_if_fail(cant_add(common_fe, selected_fe));

		Working->merge_notify (NULL, common_fe, selected_fe, MergeActionNothing);
	    } else
		Return_if_fail(merge_selected_descendant(Working, selected_fe,
						     common_fe, NULL));
	} else {
	    Return_if_fail(merge_no_descendant(Working, NULL, common_fe, NULL));
	}
    }

    return NoError;
}

static PrVoidError merge_selected_only_files(ProjectDescriptor* Selected,
					     ProjectDescriptor* /* Common */,
					     ProjectDescriptor* Working,
					     FileTable* /* selected_files */,
					     FileTable* common_files,
					     FileTable* work_files,
					     FileTable* /* selected_files_alt */,
					     FileTable* /* common_files_alt */,
					     FileTable* work_files_alt)
{
    FileEntry *selected_fe, *working_fe;

    foreach_fileent(fe_ptr, Selected) {
	selected_fe = *fe_ptr;

	if(lookup_file(selected_fe, common_files))
	    /* if its in common_files, its been handled by merge_common_files */
	    continue;

	working_fe = lookup_file(selected_fe, work_files);

	bool was_merged = Working->has_been_merged(working_fe, NULL, selected_fe);

	if(was_merged) {
	    bool remerge;
	    Return_if_fail (remerge << remerge_files (working_fe, NULL, selected_fe));

	    if (!remerge)
		continue;
	}

	if(!was_merged &&
	   !selected_fe->on_command_line() &&
	   (working_fe && !working_fe->on_command_line())) {

	    Working->merge_notify_incomplete();
	    continue;
	}

	if(!working_fe) {
	    if (alt_lookup_file (selected_fe, work_files_alt)) {
		Return_if_fail(cant_add(NULL, selected_fe));

		Working->merge_notify (NULL, NULL, selected_fe, MergeActionNothing);
	    } else
		Return_if_fail(merge_selected_only_file(Working, selected_fe, NULL, NULL));
	} else {
	    /* handled by merge_working_only_files */
	}
    }

    return NoError;
}

static PrVoidError merge_working_only_files(ProjectDescriptor* /* Selected */,
					    ProjectDescriptor* /* Common */,
					    ProjectDescriptor* Working,
					    FileTable* selected_files,
					    FileTable* common_files,
					    FileTable* /* work_files */,
					    FileTable* /* selected_files_alt */,
					    FileTable* /* common_files_alt */,
					    FileTable* /* work_files_alt */)
{
    FileEntry *work_fe, *selected_fe;

    foreach_fileent(fe_ptr, Working) {
	work_fe = *fe_ptr;

	if(lookup_file(work_fe, common_files))
	    /* handled by merge_common_files */
	    continue;

	selected_fe = lookup_file(work_fe, selected_files);

	bool was_merged = Working->has_been_merged(work_fe, NULL, selected_fe);

	if(was_merged) {
	    bool remerge;
	    Return_if_fail (remerge << remerge_files (work_fe, NULL, selected_fe));

	    if (!remerge)
		continue;
	}

	if(!was_merged &&
	   !work_fe->on_command_line() &&
	   (selected_fe && !selected_fe->on_command_line())) {

	    Working->merge_notify_incomplete();
	    continue;
	}

	if(!selected_fe) {
	    Return_if_fail(merge_work_only_file(Working, work_fe, NULL, NULL));
	} else {
	    Return_if_fail(merge_no_ancestor(Working, selected_fe, NULL, work_fe));
	}
    }

    return NoError;
}

bool any_names_differ(FileEntry* working, FileEntry* common, FileEntry* selected)
{
    return (working && common   && strcmp(working->working_path(), common->working_path()) != 0) ||
	   (working && selected && strcmp(working->working_path(), selected->working_path()) != 0) ||
	   (common  && selected && strcmp(common->working_path(), selected->working_path()) != 0);
}

static void print_filetuple(ostream& s, FileTuple tup)
{
    bool first = true;

    if (tup.work_fe) {
	if (tup.type == FileTuple::MergeTriple)
	    s << "[PW: ";
	else
	    s << "[FROM: ";

	s << tup.work_fe->working_path();

	first = false;
    }

    if (tup.com_fe) {
	if (first)
	    s << "[";
	else
	    s << ", ";

	if (tup.type == FileTuple::MergeTriple)
	    s << "PC: ";
	else
	    s << "TO: ";

	s << tup.com_fe->working_path();

	first = false;
    }

    if (tup.sel_fe) {
	if (first)
	    s << "[";
	else
	    s << ", ";

	s << "PS: " << tup.sel_fe->working_path();
    }

    s << "]";
}

ostream& __omanip_filetuple(ostream& s, FileTuple tup)
{
    if (option_file_family && any_names_differ(tup.work_fe, tup.com_fe, tup.sel_fe)) {

	print_filetuple(s, tup);

    } else {

	FileEntry* non_null;

	if(tup.work_fe)
	    non_null = tup.work_fe;
	else if (tup.sel_fe)
	    non_null = tup.sel_fe;
	else
	    non_null = tup.com_fe;

	s << squote(non_null->working_path());
    }

    return s;
}

/*
 * merge_common_descendants(S, C, W) --
 *
 *     file is present in all three versions.  In this case there are
 *     five cases:                              cmp_wc cmp_sc cmp_sw
 *	 1.  all three files are identicle      0      0      0
 *	 2.  all three files are different      1      1      1
 *	 3.  S and C are the same, W differs    1      0      1
 *	 4.  W and C are the same, S differs    0      1      1
 *	 5.  S and W are the same, C differs    1      1      0
 *     the following actions are taken:
 *	 1.  no action
 *	 2.  default merge
 *	 3.  no action
 *	 4.  default replace
 *	 5.  no action
 *
 *     Files present in PC, PS, and PW, equivalent in PC and PW, and
 *     different in PS.	 The default action is `r'.  That is, the
 *     version in PS is assumed to be most up-to-date.
 *     Files present in PC, PS, and PW, and different in all three.
 *     The default action is `m'.  There have been independent
 *     changes in both PW and PS.
 *     Otherwise PRCS does nothing.
 */
static PrVoidError merge_common_descendants(ProjectDescriptor* Merged,
					    FileEntry* selected_fe,
					    FileEntry* common_fe,
					    FileEntry* working_fe)
{
    FileType ct, st, wt;

    ct = common_fe->file_type();
    st = selected_fe->file_type();
    wt = working_fe->file_type();

    if( wt == ct && st == ct && wt == st && ct == RealFile ) {
	bool cmp_wc, cmp_sc, cmp_sw;
	RepEntry* rep_entry = Merged->repository_entry();

	cmp_sw = quick_compare (selected_fe, working_fe, working_fe->present());

	if (!cmp_sw) {
	    DEBUG ("Quick compare succeeded for "
		   << merge_tuple (working_fe, common_fe, selected_fe)
		   << ", selected == working, done");
	    return NoError;
	}

	cmp_sc = quick_compare (selected_fe, common_fe, false);

	if (!cmp_sc) {
	    DEBUG ("Quick compare succeeded for "
		   << merge_tuple (working_fe, common_fe, selected_fe)
		   << ", selected == common, done");
	    return NoError;
	}

	Return_if_fail(checkout_unkey(temp_file_selected, selected_fe, false, rep_entry, Merged));
	Return_if_fail(checkout_unkey(temp_file_common, common_fe, false, rep_entry, Merged));
	Return_if_fail(checkout_unkey(temp_file_working, working_fe, working_fe->present(),
				      rep_entry, Merged));

	Return_if_fail(cmp_sw << cmp_filenames(temp_file_selected, temp_file_working));

	if (!cmp_sw)
	    return NoError;

	Return_if_fail(cmp_sc << cmp_filenames(temp_file_selected, temp_file_common));

	if (!cmp_sc)
	    return NoError;

	cmp_wc = quick_compare (common_fe, working_fe, working_fe->present());

	if (!cmp_wc) {
	    DEBUG ("Quick compare succeeded for "
		   << merge_tuple (working_fe, common_fe, selected_fe)
		   << ", working == common");
	} else /* Possible false positive */
	    Return_if_fail(cmp_wc << cmp_filenames(temp_file_working, temp_file_common));

	if (cmp_wc) {

	    if(!option_force_resolution || option_long_format) {
		prcsoutput << "Common, selected, and working files "
			   << merge_tuple(working_fe, common_fe, selected_fe)
			   << " all differ" << dotendl;
	    }

	    Return_if_fail(merge_finish(MergeActionMerge, "dmnr", all_files_differ,
					Merged, selected_fe, common_fe, working_fe));
	} else {

	    if(!option_force_resolution || option_long_format) {
		prcsoutput << "Selected file "
			   << merge_tuple(working_fe, common_fe, selected_fe)
			   << " has been modified, working file is unmodified" << dotendl;
	    }

	    Return_if_fail(merge_finish(MergeActionReplace, "dmnr", just_selected_differs,
					Merged, selected_fe, common_fe, working_fe));
	}
    }

    return NoError;
}

/*
 * merge_working_descendant --
 *
 *     file is present in working and common version.
 *
 *     Files present in PC and PW, and absent in PS.  The default
 *     action is `d'.  That is, the files are assumed to be obsolete.
 */
static PrVoidError merge_working_descendant(ProjectDescriptor* Merged,
					    FileEntry* /* selected */,
					    FileEntry* common,
					    FileEntry* working)
{
    FileType ct, wt;

    ct = common->file_type();
    wt = working->file_type();

    if( wt == ct && ct == RealFile ) {

	if(!option_force_resolution || option_long_format)
	    prcsoutput << "No file " << merge_tuple(working, common, NULL)
		       << " in selected version, working file assumed obsolete"
		       << dotendl;

	Return_if_fail(merge_finish(MergeActionDelete, "dn", selected_file_absent,
				    Merged, NULL, common, working));
    }

    return NoError;
}

/*
 * merge_selected_descendant --
 *
 *     file is present in selected and common version.	Compare selected
 *     and common version so as to notify the user of whether it changed
 *     but default in both cases to 'n'
 *
 *     Files present in PC and absent in PW.  The default action is
 *     `n'.  That is, the files are assumed to be obsolete,
 *     regardless of their status in PS.
 */
static PrVoidError merge_selected_descendant(ProjectDescriptor* Merged,
					     FileEntry* selected,
					     FileEntry* common,
					     FileEntry* /* working */)
{
    FileType ct, st;
    const char *detail;

    ct = common->file_type();
    st = selected->file_type();

    if( st == ct && ct == RealFile ) {
	bool cmp;
	RepEntry* rep_entry = Merged->repository_entry();

	Return_if_fail(checkout_unkey(temp_file_selected, selected, false, rep_entry, Merged));
	Return_if_fail(checkout_unkey(temp_file_common, common, false, rep_entry, Merged));

	cmp = quick_compare (selected, common, false);

	if (!cmp) {
	    DEBUG ("Quick compare succeeded for "
		   << merge_tuple (NULL, common, selected)
		   << ", selected == common");
	} else /* Possible false positive */
	    Return_if_fail(cmp << cmp_filenames(temp_file_selected, temp_file_common));

	if(cmp) {

	    if(!option_force_resolution || option_long_format)
		prcsoutput << "No file " << merge_tuple(NULL, common, selected)
			<< " in working version, selected file differs from common version"
			<< dotendl;

	    detail = working_absent_sel_changed;

	} else {

	    if(!option_force_resolution || option_long_format)
		prcsoutput << "No file " << merge_tuple(NULL, common, selected)
			<< " in working version, selected file unchanged from common version"
			<< dotendl;

	    detail = working_absent_sel_unchanged;
	}

	Return_if_fail(merge_finish(MergeActionNothing, "na", detail,
				    Merged, selected, common, NULL));
    }

    return NoError;
}

/*
 * merge_selected_only_file --
 *
 *     file is only present in selected version.  Default is to add.
 *
 *     Files absent in PC and PW, and present in PS. The default
 *     action is `a'.
 *
 */
static PrVoidError merge_selected_only_file(ProjectDescriptor* Merged,
					    FileEntry* selected,
					    FileEntry* /* common */,
					    FileEntry* /* working */)
{
    if( selected->file_type() == RealFile ) {
	Return_if_fail(checkout_unkey(temp_file_selected, selected, false,
				      Merged->repository_entry(), Merged));

	if(!option_force_resolution || option_long_format)
	    prcsoutput << "No file " << merge_tuple(NULL, NULL, selected)
		       << " in working version, file not in common version" << dotendl;

	Return_if_fail(merge_finish(MergeActionAdd, "an", selected_only_detail,
				    Merged, selected, NULL, NULL));
    }

    return NoError;
}

/*
 * merge_work_only_file --
 *
 *     file is only present in working version.	 no action
 *     as per 'PRCS takes no action in other cases, and does not prompt
 *     the user.'
 */
static PrVoidError merge_work_only_file(ProjectDescriptor* /* Merged */,
					FileEntry* /* selected */,
					FileEntry* /* common */,
					FileEntry* /* working */)
{
    /* do nothing */
    return NoError;
}

/*
 * merge_no_descendant --
 *
 *     file is present only in common version.	no action
 *     as per 'PRCS takes no action in other cases, and does not prompt
 *     the user.'
 */
static PrVoidError merge_no_descendant(ProjectDescriptor* /* Merged */,
				FileEntry*, FileEntry*, FileEntry*)
{
    /* do nothing */
    return NoError;
}

/*
 * merge_no_ancestor --
 *
 *     file is present in selected and working versions, but not
 *     common version.
 *     Files absent in PC, and different in PW and PS.	The default
 *     action is `m'
 */
static PrVoidError merge_no_ancestor(ProjectDescriptor* Merged,
			      FileEntry* selected,
			      FileEntry* /* common */,
			      FileEntry* working)
{
    FileType st, wt;

    st = selected->file_type();
    wt = working->file_type();

    if( wt == st && wt == RealFile ) {
	bool cmp;
	RepEntry* rep_entry = Merged->repository_entry();

	cmp = quick_compare (selected, working, working->present());

	if (!cmp) {
	    DEBUG ("Quick compare succeeded for "
		   << merge_tuple (working, NULL, selected)
		   << ", selected == working, done");
	    return NoError;
	}

	Return_if_fail(checkout_unkey(temp_file_selected, selected, false, rep_entry, Merged));
	Return_if_fail(fs_zero_file(temp_file_common));
	Return_if_fail(checkout_unkey(temp_file_working, working, working->present(), rep_entry, Merged));

	if (!cmp)
	    Return_if_fail(cmp << cmp_filenames(temp_file_working, temp_file_selected));

	if (!cmp) {
	    return NoError;
	} else {

	    if(!option_force_resolution || option_long_format)
		prcsoutput << "Working and selected files " << merge_tuple(working, NULL, selected)
			   << " differ, no ancestor, merge on the empty file" << dotendl;

	    Return_if_fail(merge_finish(MergeActionMerge, "dmnr", no_ancestor_detail,
					Merged, selected, NULL, working));
	}
    }

    return NoError;
}

static PrBoolError cmp_filenames(const char* file1, const char* file2)
{
    int one, two;

    If_fail(one << Err_open(file1, O_RDONLY)) {
	throw prcserror << "Open failed on " << squote(file1) << perror;
    }

    If_fail(two << Err_open(file2, O_RDONLY)) {
	throw prcserror << "Open failed on " << squote(file2) << perror;
    }

    bool diffs;

    If_fail(diffs << cmp_fds(one, two)) {
	close(one);
	close(two);
	throw prcserror << "Read failed while comparing " << squote(file1)
			<< " and " << squote(file2) << perror;
    }

    close(one);
    close(two);

    return diffs;
}

static PrVoidError checkout_unkey(const char* filename, FileEntry* fe,
				  bool working, RepEntry* rep_entry,
				  ProjectDescriptor *project)
{
    const char* rcsversion, *versionfile, *working_path;

    Return_if_fail(fe->initialize_descriptor(rep_entry, false, false));

    if(working) {
	Return_if_fail(fe->get_file_sys_info());
    } else {
	Return_if_fail(fe->get_repository_info(rep_entry));
    }

    rcsversion = fe->descriptor_version_number();
    versionfile = fe->full_descriptor_name();
    working_path = fe->working_path();

    if (working) {
	if(fe->keyword_sub()) {
	    Return_if_fail( setkeys(working_path,
				    filename,
				    fe,
				    project,
				    Unsetkeys) );
	} else {
	    /* Can't symlink here, because gdiff's output is written
	     * over it.  Its bad when you start merging files onto
	     * themselves. */
	    Return_if_fail(fs_copy_filename(working_path, filename));
	}
    } else {
	Return_if_fail(VC_checkout_file(filename, rcsversion, versionfile));
    }

    return NoError;
}

static void print_merge_help()
{
    prcsoutput << "Valid options are:" << prcsendl
	    << "n -- Do nothing" << dotendl
	    << "m -- Merge selected and working files" << dotendl
	    << "r -- Replace working file with selected file" << dotendl
	    << "d -- Delete working file" << dotendl
	    << "a -- Add selected file to working version" << dotendl
	    << "h -- Explain this condition" << dotendl
	    << "v -- View diffs" << dotendl
	    << "! -- Take the default and do not offer this query again" << dotendl
	    << "q -- " << default_fail_query_message << dotendl;
}

PrVoidError view_sc_diffs(FileEntry* selected, FileEntry* common, ProjectDescriptor *project)
{
    Return_if_fail(diff_pair(common, selected,
			     temp_file_common,
			     temp_file_selected,
			     "common",
			     "selected",
			     project));
    return NoError;
}

PrVoidError view_wc_diffs(FileEntry* working, FileEntry* common, ProjectDescriptor *project)
{
    Return_if_fail(diff_pair(common, working,
			     temp_file_common,
			     temp_file_working,
			     "common",
			     "working",
			     project));
    return NoError;
}

PrVoidError view_sw_diffs(FileEntry* working, FileEntry* selected, ProjectDescriptor *project)
{
    Return_if_fail(diff_pair(selected, working,
			     temp_file_selected,
			     temp_file_working,
			     "selected",
			     "working",
			     project));
    return NoError;
}

static PrVoidError view_merge_diffs(FileEntry* selected,
				    FileEntry* common,
				    FileEntry* working,
				    ProjectDescriptor *project)
{
    while(true) {
	if(selected && common)
	    prcsquery << option('s', "View selected - common diffs");

	if(working && common)
	    prcsquery << option('w', "View working - common diffs");

	if(selected && working)
	    prcsquery << option('d', "View working - selected diffs");

	prcsquery << defopt('l', "Leave this menu") << query("Please select");

	char c;

	Return_if_fail(c << prcsquery.result());

	switch(c) {
	case 's':
	    Return_if_fail(view_sc_diffs(selected, common, project));
	    break;
	case 'w':
	    Return_if_fail(view_wc_diffs(working, common, project));
	    break;
	case 'd':
	    Return_if_fail(view_sw_diffs(working, selected, project));
	    break;
	case 'l':
	    return NoError;;
	}
    }
}

static PrVoidError merge_finish(MergeAction def_action,
				const char* valid,
				const char* detail,
				ProjectDescriptor* Merged,
				FileEntry* selected,
				FileEntry* common,
				FileEntry* working)
{
    MergeAction action;

    if(option_report_actions) {
	Merged->merge_notify(working, common, selected, def_action);

	switch(def_action) {
	case MergeActionMerge:
	    prcsoutput << "Merge file:   " << merge_tuple(working, common, selected)
		       << prcsendl;
	    break;
	case MergeActionReplace:
	    prcsoutput << "Replace file: " << merge_tuple(working, common, selected)
		       << prcsendl;
	    break;
	case MergeActionAdd:
	    prcsoutput << "Add file:     " << merge_tuple(working, common, selected)
		       << prcsendl;
	    break;
	case MergeActionNothing:
	    return NoError;
	case MergeActionDelete:
	    prcsoutput << "Delete file:  " << merge_tuple(working, common, selected)
		       << prcsendl;
	    break;
	}

	return NoError;
    } else if(option_force_resolution) {
	action = def_action;
    } else {
	while(true) {
	    static bool bang_flag = false;
	    char c;

	    if(bang_flag) {
		action = def_action;
		break;
	    }

	    Dstring query;

	    query.sprintf("Please select(%svh!?)[%c] ", valid, (char)def_action);

	    re_query_message = query.cast();
	    re_query_len = query.length();

	    re_query();

	    c = get_user_char();

	    if(c == 0) fprintf(stdout, "\n");

	    if(c == 0 || c == 'q') return PrVoidError(UserAbort);

	    if(c == '?')
		print_merge_help();
	    else if(c == 'h')
		prcsoutput << detail << prcsendl;
	    else if(c == 'v')
		Return_if_fail(view_merge_diffs(selected, common, working, Merged));
	    else if(c == '\n') {
		action = def_action;
		break;
	    } else if(c == '!') {
		action = def_action;
		bang_flag = true;
		break;
	    } else if(strchr(valid, c) != NULL) {
		action = (MergeAction)c;
		break;
	    }
	}
    }

    re_query_message = NULL;

    switch(action) {
    case MergeActionMerge:
	bool conflict;

	Return_if_fail(conflict << merge_files(Merged, selected, common, working));

	prcsoutput << "Merge file:   " << merge_tuple(working, common, selected);

	if(conflict) {
	    prcsoutput << ", conflicts created" << prcsendl;

	    static bool edit_conflicts = false;
	    static const char* conflict_editor;

	    if (!edit_conflicts) {
		conflict_editor = getenv ("PRCS_CONFLICT_EDITOR");
		edit_conflicts = true;
	    }

	    if (conflict_editor) {
		static SystemCommand* edit_cmd = NULL;

		if (!edit_cmd)
		    edit_cmd = new SystemCommand(conflict_editor);

		ArgList* args;

		Return_if_fail(args << edit_cmd->new_arg_list());

		args->append(*working->working_path());

		Return_if_fail(edit_cmd->open_stdout());
		/* RETURN VALUE OF EDITOR IGNORED. */
	    }
	} else {
	    prcsoutput << prcsendl;
	}

	break;
    case MergeActionReplace:
	prcsoutput << "Replace file: " << merge_tuple(working, common, selected) << prcsendl;
	Return_if_fail(replace_file(Merged, selected, common, working));
	break;
    case MergeActionAdd:
	prcsoutput << "Add file:     " << merge_tuple(working, common, selected) << prcsendl;
	Return_if_fail(add_file(Merged, selected, common, working));
	break;
    case MergeActionDelete:
	prcsoutput << "Delete file:  " << merge_tuple(working, common, selected) << prcsendl;
	Return_if_fail(delete_file(Merged, selected, common, working));
	break;
    case MergeActionNothing:
	break;
    }

    Merged->merge_notify(working, common, selected, action);
    return NoError;
}

static PrBoolError merge_files(ProjectDescriptor* Merged,
			       FileEntry* selected,
			       FileEntry* common,
			       FileEntry* working)
{
    ArgList *args;
    Dstring slabel, clabel, wlabel;
    int ret;

    /* Take the selected version history here, because the working
     * version is not always guaranteed to have one. */
    working->set_descriptor_name(selected->descriptor_name());
    working->set_version_number(selected->descriptor_version_number());
    working->set_file_mode(selected->file_mode());

    If_fail(working->stat())
	throw prcserror << "Stat failed on "
			<< squote(working->working_path()) << perror;

    Return_if_fail(args << gdiff3_command.new_arg_list());

    /*
     * If merging a no_ancestor set of files, common will be null, so
     * test the value of the file entry before formatting the label.
     */
    ASSERT(selected, "illegal file entry, only common may be null");
    ASSERT(working,  "illegal file entry, only common may be null");

    format_diff_label(selected, selected->working_path(), selected_format, &slabel, Merged);
    format_diff_label(working, working->working_path(), working_format, &wlabel, Merged);

    if(common)
	format_diff_label(common, common->working_path(), common_format, &clabel, Merged);
    else
	clabel.sprintf("-L*** empty file ***");

    args->append("-maE");
    args->append(wlabel);
    args->append(temp_file_working);
    args->append(clabel);
    args->append(temp_file_common);
    args->append(slabel);
    args->append(temp_file_selected);

    Return_if_fail(make_obsolete(working->working_path()));

    WriteableFile outfile;

    Return_if_fail(outfile.open(working->working_path()));

    Return_if_fail(gdiff3_command.open(true, false));

    Return_if_fail(outfile.copy(gdiff3_command.standard_out()));

    Return_if_fail(outfile.close());

    Return_if_fail(ret << gdiff3_command.close());

    if(ret > 1)
	throw prcserror << "gdiff3 command failed" << dotendl;

    Return_if_fail(working->chmod(working->stat_file_mode()));

    return (bool)ret;
}

static PrVoidError replace_file(ProjectDescriptor* /* Merged */,
				FileEntry* selected,
				FileEntry* /* common */,
				FileEntry* working)
{
    working->set_descriptor_name(selected->descriptor_name());
    working->set_version_number(selected->descriptor_version_number());
    working->set_file_mode(selected->file_mode());

    If_fail(working->stat())
	throw prcserror << "Stat failed on "
			<< squote(working->working_path()) << perror;

    Return_if_fail(make_obsolete(working->working_path()));

    WriteableFile outfile;

    Return_if_fail(outfile.open(working->working_path()));

    Return_if_fail(outfile.copy(temp_file_selected));

    Return_if_fail(outfile.close());

    Return_if_fail(working->chmod(working->stat_file_mode()));

    return NoError;
}

static PrVoidError make_obsolete(const char* work_path)
{
    char *lastslash = strrchr(work_path, '/');
    char buf[20];
    int len, tried = 0;
    Dstring obs(work_path);

    if(lastslash != NULL)
	len = (lastslash - work_path) + 1;
    else
	len = 0;

    do {
	obs.truncate(len);

	if(tried == 0)
	    buf[0] = '\0';
	else
	    sprintf(buf, "%d", tried++);

	obs.sprintfa("obsolete%s", buf);
    } while(fs_file_exists(obs) && !fs_is_directory(obs));

    if(!fs_file_exists(obs)) {
	If_fail(Err_mkdir(obs, 0777))
	    throw prcserror << "Mkdir failed on directory " << squote(obs) << perror;
    }

    obs.append('/');
    obs.append((const char*)work_path + len);

    Return_if_fail(fs_move_filename(work_path, obs));

    if(option_long_format)
	prcsinfo << "Copied working file " << squote(work_path)
		 << " to obsolete subdir " << squote(obs) << dotendl;

    return NoError;
}

static PrVoidError delete_file(ProjectDescriptor* Merged,
			       FileEntry* /* selected */,
			       FileEntry* /* common */,
			       FileEntry* working)
{
    Return_if_fail(make_obsolete(working->working_path()));

    If_fail(Err_unlink(working->working_path()))
       prcswarning << "Unlink failed on "
		   << squote(working->working_path()) << perror;

    Merged->delete_file(working);

    return NoError;
}

static PrVoidError add_file(ProjectDescriptor* Merged,
			    FileEntry* selected,
			    FileEntry*, FileEntry*)
{
    static bool do_once = true;

    if(do_once) {
	Merged->append_files_data("\n");
	do_once = false;
    }

    Merged->append_file(selected);

    Return_if_fail(make_subdirectories(selected->working_path()));

    Return_if_fail(fs_move_filename(temp_file_selected, selected->working_path()));

    Return_if_fail(selected->chmod(0));

    return NoError;
}

static PrVoidError merge_cleanup_handler(void* data, bool)
{
    if(option_report_actions)
	return NoError;

    ProjectDescriptor* Merged = ((ProjectDescriptor*)data);

    Merged->merge_finish();

    Return_if_fail(Merged->write_project_file(cmd_root_project_file_path));

    return NoError;
}

MergeCandidate::MergeCandidate (ProjectVersionData *data0, bool is_working0)
    :_data(data0),
     _is_working(is_working0),
     _common(NULL),
     _common_distance(-1) { }
