/* -*-Mode: C++;-*-
 * PRCS - The Project Revision Control System
 * Copyright (C) 1997  Josh MacDonald
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: package.cc 1.15.1.2.1.8.1.8.1.7.1.2 Sat, 11 Jan 1997 23:19:49 -0800 jmacd $
 */

extern "C" {
#include <sys/wait.h>
#include <signal.h>
}

#include "prcs.h"
#include "syscmd.h"
#include "misc.h"
#include "repository.h"
#include "projdesc.h"
#include "system.h"
#include "vc.h"

const char gzip_magic_1 = '\037';
const char gzip_magic_2 = '\213';
const char magic_number_1 = '\200';     /* tar file contains PROJECT/files */
const char magic_number_2 = '\352';
const char new_magic_number_1 = '\201'; /* tar file contains files, allows easy renaming */
const char new_magic_number_2 = '\353';
const int header_length = 256;
const mode_t package_file_mode = 0666;

static PrVoidError package_project(const char* project_name, const char* package_file);

PrPrcsExitStatusError package_command()
{
    Return_if_fail(package_project(cmd_root_project_name, cmd_filenames_given[0]));

    return ExitSuccess;
}

static PrVoidError package_project(const char* project_name, const char* package_file)
{
    ArgList *args;
    int outfd = STDOUT_FILENO;
    char buf[header_length];
    int pipefd[2], pid = -1, status;
    const char* rep_path;

    Return_if_fail(args << tar_command.new_arg_list());

    If_fail(rep_path << Rep_guess_repository_path())
	throw prcserror << "Please set your PRCS_REPOSITORY environment variable" << dotendl;

    if(!fs_is_directory(rep_path))
	throw prcserror << "Repository " << squote(rep_path) << " does not exist" << dotendl;

    Dstring rep_entry_path(rep_path);
    rep_entry_path.append('/');
    rep_entry_path.append(project_name);

    if(!fs_is_directory(rep_entry_path))
	throw prcserror << "Repository entry " << squote(rep_entry_path) << " does not exist" << dotendl;

    if(strcmp(package_file, "-") != 0) {
	If_fail(outfd << Err_open(package_file, O_WRONLY|O_CREAT|O_TRUNC, package_file_mode))
	    throw prcserror << "Can't open " << squote(package_file)
		            << " for writing" << perror;
    }

    args->append("cf");
    args->append("-");
    args->append("-C");
    args->append(rep_entry_path);
    args->append(".");

    Return_if_fail(tar_command.open(true, false));

    memset(buf, 0, header_length);
    buf[0] = new_magic_number_1;
    buf[1] = new_magic_number_2;
    sprintf(buf + 2, "%s", project_name);

    if(option_package_compress) {
	Return_if_fail(gzip_command.init());

	if(pipe(pipefd) != 0) {
	    throw prcserror << "Pipe failed, can't compress data" << perror;
	} else if((pid = fork()) < 0) {
	    throw prcserror << "Fork failed, can't compress data" << perror;
	} else if (pid == 0) {
	    dup2(pipefd[PIPE_READ_FD], STDIN_FILENO);
	    dup2(outfd, STDOUT_FILENO);
	    close(pipefd[PIPE_READ_FD]);
	    close(pipefd[PIPE_WRITE_FD]);

	    execl(gzip_command.path(), "-", "-f", NULL);
	    abort_child(gzip_command.path());
	} else {
	    dup2(pipefd[PIPE_WRITE_FD], outfd);
	    close(pipefd[PIPE_READ_FD]);
	    close(pipefd[PIPE_WRITE_FD]);
	}
    }

    If_fail(Err_write(outfd, buf, header_length))
	throw prcserror << "Write failed writing package file" << perror;

    If_fail(write_fds(fileno(tar_command.standard_out()), outfd))
	throw prcserror << "Write failed writing package file" << perror;

    If_fail(Err_close(outfd))
	throw prcserror << "Write failed writing package file" << perror;

    Return_if_fail_if_ne(tar_command.close(), 0) {
	unlink(package_file);
	throw prcserror << "System command tar failed" << dotendl;
    }

    if(option_package_compress) {
	If_fail(status << Err_waitpid(pid))
	    throw prcserror << "Waitpid failed on pid " << pid << perror;

	if(!WIFEXITED(status) || WEXITSTATUS(status) != 0)
	    throw prcserror << "System command gzip failed" << dotendl;
    }

    return NoError;
}

static PrVoidError invalid()
{
    throw prcserror << squote(cmd_root_project_name)
		    << " is not a valid package file" << dotendl;
}

static PrVoidError feed_pipe(const char* header_buf, int from_fd, int to_fd)
{
    If_fail(Err_write(to_fd, header_buf, header_length))
	throw prcserror << "Write failed writing to gzip pipe" << perror;

    If_fail(write_fds(from_fd, to_fd))
	throw prcserror << "Write failed writing to gzip pipe" << perror;

    return NoError;
}

PrPrcsExitStatusError unpackage_command()
{
    ArgList *args;
    int fd, zipoutpipefd[2], zipinpipefd[2], zippid = 0, writepid = 0;
    Dstring projname;
    char buf[header_length];
    const char* rep_path;
    bool new_format = false;
    bool do_rename = false;

    if(strcmp(cmd_root_project_name, "-") != 0) {
	If_fail(fd << Err_open(cmd_root_project_full_name, O_RDONLY))
	    throw prcserror << "Open failed on package file" << perror;
    } else {
	fd = STDIN_FILENO;
    }

    if (cmd_filenames_count > 1)
	throw prcserror << "Too many file-or-dir options" << dotendl;

    If_fail(rep_path << Rep_guess_repository_path())
	throw prcserror << "Please set your PRCS_REPOSITORY environment variable" << dotendl;

    if(!fs_is_directory(rep_path))
	throw prcserror << "Repository " << squote(rep_path) << " does not exist" << dotendl;

    If_fail(Err_read_expect(fd, buf, header_length)) {
	if (errno)
	    throw prcserror << "Read error on package file" << perror;
	else
	    return invalid();
    }

    if(buf[0] == magic_number_1 && buf[1] == magic_number_2) {
	option_package_compress = false;
    } else if(buf[0] == new_magic_number_1 && buf[1] == new_magic_number_2) {
	option_package_compress = false;
	new_format = true;
    } else if(buf[0] == gzip_magic_1 && buf[1] == gzip_magic_2) {
	option_package_compress = true;
    } else {
	return invalid();
    }

    if(option_package_compress) {
	Return_if_fail(gzip_command.init());

	if(pipe(zipoutpipefd) != 0 || pipe(zipinpipefd) != 0) {
	    throw prcserror << "Pipe failed, can't compress data" << perror;
	} else if((zippid = fork()) < 0) {
	    throw prcserror << "Fork failed, can't compress data" << perror;
	} else if (zippid == 0) {
	    dup2(zipoutpipefd[PIPE_WRITE_FD], STDOUT_FILENO);
	    dup2(zipinpipefd[PIPE_READ_FD], STDIN_FILENO);

	    close(zipoutpipefd[PIPE_READ_FD]);
	    close(zipoutpipefd[PIPE_WRITE_FD]);
	    close(zipinpipefd[PIPE_READ_FD]);
	    close(zipinpipefd[PIPE_WRITE_FD]);

	    execl(gzip_command.path(), "-", "-d","-f", NULL);
	    abort_child(gzip_command.path());
	} else if((writepid = fork()) < 0) {
	    throw prcserror << "Fork failed, can't compress data" << perror;
	} else if(writepid == 0) {
	    int exit_val = 0;

            close(zipinpipefd[PIPE_READ_FD]);

	    If_fail(feed_pipe(buf, fd, zipinpipefd[PIPE_WRITE_FD]))
		exit_val = 1;

	    close(zipinpipefd[PIPE_WRITE_FD]);

	    _exit(exit_val);
	} else {

	    dup2(zipoutpipefd[PIPE_READ_FD], fd);
	    close(zipoutpipefd[PIPE_READ_FD]);
	    close(zipoutpipefd[PIPE_WRITE_FD]);
	    close(zipinpipefd[PIPE_READ_FD]);
	    close(zipinpipefd[PIPE_WRITE_FD]);

	    If_fail(Err_read_expect(fd, buf, header_length))
		return invalid();

	    if (buf[0] == magic_number_1 || buf[1] == magic_number_2) {

	    } else if (buf[0] == new_magic_number_1 || buf[1] == new_magic_number_2) {
		new_format = true;
	    } else {
		return invalid();
	    }
	}
    }

    if (cmd_filenames_count == 1) {
	if (!new_format)
	    throw prcserror << "Cannot rename project in package file created by PRCS versions "
		"older than 1.1.0" << dotendl;

	if (VC_check_token_match(cmd_filenames_given[0], "label") <= 0)
	    throw prcserror << "Illegal project name "
			    << squote(cmd_filenames_given[0]) << dotendl;

	if (strcmp (projname, buf + 2) != 0)
	    do_rename = true;

	projname.assign (cmd_filenames_given[0]);

	prcsinfo << "Unpackage project " << squote(projname)
		 << ", formerly named " << squote(buf + 2) << dotendl;
    } else {
	projname.assign (buf + 2);

	prcsinfo << "Unpackage project " << squote(projname) << dotendl;
    }

    Return_if_fail(args << tar_command.new_arg_list());

    Dstring entry_path(rep_path);
    entry_path.append('/');
    entry_path.append(projname);

    if(fs_is_directory(entry_path)) {
	prcsquery << "Project " << squote(projname) << " already exists.  "
		  << force("Replacing")
		  << report("Replace")
		  << optfail('n')
		  << defopt('y', "Delete project in repository")
		  << query("Replace");

	Return_if_fail(prcsquery.result());

	if(option_report_actions)
	    return ExitSuccess;

	If_fail(fs_nuke_file(entry_path)) {
	    throw prcserror << "Failed removing old repository entry "
			     << squote(entry_path) << dotendl;
	}
    }

    if(option_report_actions)
	return ExitSuccess;

    const char* change_to_dir = new_format ? entry_path : rep_path;

    if (new_format) {
	If_fail (Err_mkdir (change_to_dir, 0777))
	    throw prcserror << "Mkdir " << squote(change_to_dir) << " failed" << perror;
    }

    If_fail(change_cwd(change_to_dir))
	/* Sun's tar doesn't accept -C for 'x' operations, pretty lame huh. */
	throw prcserror << "Chdir " << squote(change_to_dir) << " failed" << perror;

    args->append("-xpf");
    args->append("-");

    Return_if_fail_if_ne(tar_command.open_filein(fd, false), 0)
	throw prcserror << "System command tar failed" << dotendl;

    int status;

    if(option_package_compress) {
	If_fail(status << Err_waitpid(writepid))
	    throw prcserror << "Waitpid failed on pid " << writepid << perror;

	if(!WIFEXITED(status) || WEXITSTATUS(status) != 0)
	    throw prcserror << "Pipe reader exited with non-zero status" << dotendl;

	If_fail(status << Err_waitpid(zippid))
	    throw prcserror << "Waitpid failed on pid " << zippid << perror;

	if(WIFSIGNALED(status) && WTERMSIG(status) == SIGPIPE)
	    throw prcserror << "Child gzip process was terminated with SIGPIPE" << dotendl;

	if(!WIFEXITED(status) || WEXITSTATUS(status) != 0)
	    throw prcserror << "Child gzip process returns non-zero status" << dotendl;
    }

    if (do_rename) {
	Dstring old_name (buf+2);

	old_name.append (".prj,v");
	projname.append (".prj,v");

	If_fail (Err_rename (old_name, projname))
	    throw prcserror << "Rename " << squote(old_name) << " to " << squote(projname)
			    << " failed" << perror;
    }

    return ExitSuccess;
}

PrVoidError save_package()
{
    Dstring new_package;
    bool compress_it = true;

    new_package.append(cmd_root_project_name);
    new_package.append(".pkg");

    prcsquery << "Enter a package name"
	      << definput(new_package)
	      << string_query("");

    const char* ans;

    Return_if_fail(ans << prcsquery.string_result());

    prcsquery << "Packaging repository.  "
	      << force("Compressing")
	      << report("Compress")
	      << option('n', "Don't compress the package")
	      << defopt('y', "Compress the package")
	      << query("Compress");

    char c;

    Return_if_fail(c << prcsquery.result());

    compress_it = (c == 'y');

    option_package_compress = compress_it;

    Return_if_fail(package_project(cmd_root_project_name, ans));

    return NoError;
}
